"""Test the helicopyter module."""

from cdktf import TerraformLocal, TerraformOutput, TerraformVariable
from cdktf_cdktf_provider_null.resource import Resource as NullResource
from pytest import raises

from helicopyter import HeliStack


def test_helistack() -> None:
    """The class must instantiate and provide the cona attribute and provide and push methods."""
    stack = HeliStack('foo')
    assert stack.cona == 'foo'
    assert callable(stack.provide)
    assert callable(stack.push)


def test_push_id() -> None:
    """Within a given Element such as the NullResource, the id_ must be unique."""
    stack = HeliStack('foo')
    my_first_null = stack.push(NullResource, 'bar')
    assert isinstance(my_first_null, NullResource)

    my_second_null = stack.push(NullResource, 'baz')
    assert isinstance(my_second_null, NullResource)

    with raises(RuntimeError):
        stack.push(NullResource, 'bar')


def test_push_provider() -> None:
    """The same id_ must be allowed for different Elements."""
    stack = HeliStack('foo')
    stack.push(NullResource, 'bar')
    stack.push(TerraformLocal, 'bar', 'bar')
    stack.push(TerraformOutput, 'bar', value='bar')
    stack.push(TerraformVariable, 'bar')
    stack.to_terraform()



"""Grant access to people in the Biobuddies GitHub organization."""

from cdktf_cdktf_provider_github.membership import Membership

from helicopyter import HeliStack

buddies = {
    'christopher.covington': ('covracer', 'admin'),
    'darren.pham': ('darpham', 'admin'),
    'duncan.tormey': ('DuncanTormey', 'admin'),
    'james.braza': ('jamesbraza', 'admin'),
    'matt.fowler': ('mattefowler', 'admin'),
}


def synth(stack: HeliStack) -> None:
    stack.provide('github')
    for firstname_dot_lastname, (username, role) in buddies.items():
        stack.push(
            Membership, firstname_dot_lastname.replace('.', '_'), username=username, role=role
        )



"""Codify Allowedflare infrastructure."""

from ast import literal_eval
from os import environ

from cdktf_cdktf_provider_cloudflare.access_application import AccessApplication
from cdktf_cdktf_provider_cloudflare.access_identity_provider import AccessIdentityProvider
from cdktf_cdktf_provider_cloudflare.access_policy import AccessPolicy, AccessPolicyInclude
from cdktf_cdktf_provider_cloudflare.worker_route import WorkerRoute

from helicopyter import HeliStack


def synth(stack: HeliStack) -> None:
    stack.provide('cloudflare')

    # If this was a private repository, I'd probably set these variables using string literals
    account_id = environ['CLOUDFLARE_ACCOUNT_ID']
    emails = literal_eval(environ['ALLOWEDFLARE_EMAILS'])
    private_domain = environ['ALLOWEDFLARE_PRIVATE_DOMAIN']
    zone_id = environ['CLOUDFLARE_ZONE_ID']

    stack.push(
        AccessIdentityProvider,
        'this',
        account_id=account_id,
        name='One-time PIN',
        type='onetimepin',
    )

    application = stack.push(
        AccessApplication,
        'this',
        domain=f'*.{private_domain}',
        session_duration='24h',
        skip_interstitial=True,
    )

    stack.push(
        AccessPolicy,
        'email-in',
        name='email-in',
        application_id=application.id,
        decision='allow',
        precedence=1,
        include=[AccessPolicyInclude(email=emails)],
    )

    stack.push(
        AccessPolicy,
        'service-in',
        name='service-in',
        application_id=application.id,
        decision='non_identity',
        precedence=2,
        include=[AccessPolicyInclude(any_valid_service_token=True)],
    )

    stack.push(
        WorkerRoute,
        'this',
        pattern=f'*.{private_domain}/x/*',
        script_name='allowedflare-proxy',
        zone_id=zone_id,
    )



"""Demonstrate a simple HeliStack synth function using CDKTF constructs."""

from cdktf import LocalExecProvisioner, TerraformLocal, TerraformOutput, TerraformVariable
from cdktf_cdktf_provider_null.resource import Resource as NullResource

from helicopyter import HeliStack


def synth(stack: HeliStack) -> None:
    """
    Accept the Git hASH (GASH) as a variable and output it.

    Also infer the ENVIronment (ENVI) from the workspace and echo it to standard output.
    """
    stack.push(TerraformLocal, 'cona', stack.cona)
    stack.push(TerraformLocal, 'envi', '${terraform.workspace}')

    stack.push(
        NullResource,
        'this',
        provisioners=[
            LocalExecProvisioner(
                command='echo $envi', environment={'envi': '${local.envi}'}, type='local-exec'
            )
        ],
    )
    gash = stack.push(TerraformVariable, 'gash', type='string')
    stack.push(TerraformOutput, 'gash', value=gash.to_string())



from dataclasses import dataclass
from functools import partial
from typing import ClassVar, TypedDict


class Registry:
    def __init__(self, cls):
        self.cls = cls
        self.instances: dict[str, cls] = {}

    def __getattr__(self, name):
        if name in self.instances:
            return self.instances[name]
        return partial(self.cls, self, name)

    def __str__(self):
        return self.cls._str


registries: dict[str, Registry] = {}


@dataclass
class RegisteredDataclass:
    _str: ClassVar[str]
    _registry: Registry
    _name: str

    def __post_init__(self):
        print(f'initializing {self._registry}.{self._name}')
        #registries[]
        self._registry.instances[self._name] = self


@dataclass
class NullResource(RegisteredDataclass):
    """
    The `null_resource` resource implements the standard resource lifecycle but takes no further action.
    On Terraform 1.4 and later, use the [terraform_data resource
    type](https://developer.hashicorp.com/terraform/language/resources/terraform-data) instead.  The
    `triggers` argument allows specifying an arbitrary set of values that, when changed, will cause
    the resource to be replaced.
    """

    _str: ClassVar[str] = 'null_resource'
    _registry: Registry
    _name: str
    #: A map of arbitrary strings that, when changed, will force the null resource to be replaced,
    #: re-running any associated provisioners.
    triggers: dict[str] | None = None

    @property
    def id(self) -> str:
        """This is set to a random value at create time."""


null_resource = Registry(NullResource)





"""Configure GitHub repositories."""

from cdktf_cdktf_provider_github.repository import Repository

from helicopyter import HeliStack

repositories = {
    'airdjang': ('Airflow + Django', ['airflow', 'django', 'python']),
    'allowedflare': ('Intranet connectivity for Django and more', ['django', 'python']),
    'helicopyter': ('Python-defined infrastructure', ['ansible', 'cdktf', 'python', 'terraform']),
}


def synth(stack: HeliStack) -> None:
    stack.provide('github')

    for name, (description, topics) in repositories.items():
        stack.push(
            Repository,
            name,
            import_id=name,
            allow_auto_merge=True,
            allow_merge_commit=False,
            allow_rebase_merge=True,
            allow_squash_merge=True,
            allow_update_branch=True,
            delete_branch_on_merge=True,
            description=description,
            has_downloads=False,
            has_issues=True,
            has_projects=False,
            has_wiki=False,
            name=name,
            # Merge commits aren't expected but just in case
            merge_commit_message='PR_BODY',
            merge_commit_title='PR_TITLE',
            # This might be the most important setting: it copies the the body (top comment) of the
            # Pull Request (PR) on github.com into the git commit message. The default is to list
            # the commit titles which are often much less helpful:
            # "add feature; fix test; really fix test; satisfy linter"
            squash_merge_commit_message='PR_BODY',
            squash_merge_commit_title='PR_TITLE',
            topics=topics,
        )



"""Generate JSON which Terraform can use from Python."""

from collections.abc import Iterable
from importlib import import_module
from json import dump
from pathlib import Path
from subprocess import check_output
from typing import Any, TypeVar

from cdktf import App, TerraformElement, TerraformStack
from constructs import Construct, Node
from tap import Tap


class HeliStack(TerraformStack):
    def __init__(self, cona: str) -> None:
        # Something is automatically creating outdir, which is cdktf.out by default
        super().__init__(App(outdir='.'), cona)

        self.cona = cona
        self.imports: dict[str, str] = {}  # {to_0: id_0, to_1, id_1, ...}
        self._scopes: dict[str, Construct] = {}

    def _allocate_logical_id(self, element: Node | TerraformElement) -> str:
        if isinstance(element, Node):
            # Mostly for mypy. Patches to support AWS CDK welcome.
            raise TypeError('AWS CDK unsupported; please use CDKTF')
        return element.node.id

    def provide(self, name: str) -> type[TerraformElement]:
        """
        Return a Provider class instance given its short name.

        Example usage:
        stack.provide('cloudflare')
        """
        module = import_module(f'cdktf_cdktf_provider_{name}.provider')
        return getattr(module, f'{name.title()}Provider')(self, 'this')

    E = TypeVar('E', bound=TerraformElement)

    def push(
        self,
        Element: type[E],  # noqa: N803
        id_: str,
        *args: Any,  # noqa: ANN401
        import_id: str = '',
        **kwargs: Any,  # noqa: ANN401
    ) -> E:
        """
        Add Data or a Resource to the stack and return it.

        Example usage:
        from cdktf_cdktf_provider_cloudflare.access_application import AccessApplication
        stack.push(AccessApplication, 'mydomain-wildcard', domain='*.mydomain.com')
        """
        if Element.__module__ == 'cdktf':
            scope_name = Element.__name__.lower().replace('terraform', '')
        else:
            scope_name = Element.__module__.replace('cdktf_cdktf_provider_', '').replace('.', '_')
        if scope_name not in self._scopes:
            self._scopes[scope_name] = Construct(self, scope_name)

        print(f'Pushing {scope_name}.{id_}')
        element = Element(self._scopes[scope_name], id_, *args, **kwargs)
        if import_id:
            self.imports[
                f"{Element.__module__.replace('cdktf_cdktf_provider_', '').replace('.', '_')}.{id_}"
            ] = import_id
        return element


# ruff: noqa: T201
def multisynth(
    all_or_conas_or_paths: Iterable[str], *, hashicorp_configuration_language: bool = False
) -> None:
    if not all_or_conas_or_paths:
        print('No codenames specified. Doing nothing.')
        return

    top_directory = Path(__file__).parent

    if 'all' in all_or_conas_or_paths or 'helicopyter.py' in all_or_conas_or_paths:
        conas_or_paths = {
            file.parent.parent.name
            for file in (top_directory / 'deploys').glob('**/terraform/main.py')
        }
    else:
        conas_or_paths = set(all_or_conas_or_paths)

    for cona_or_path in conas_or_paths:
        path_to_check = Path(cona_or_path)
        if (
            path_to_check.exists()
            and path_to_check.name == 'main.py'
            and path_to_check.parent.name == 'terraform'
        ):
            cona = path_to_check.parent.parent.name
        else:
            cona = cona_or_path
        relative_path = (
            f'deploys/{cona}/terraform/main.tf{"" if hashicorp_configuration_language else ".json"}'
        )
        print(f'Generating {relative_path}')
        module = import_module(f'deploys.{cona}.terraform.main')
        stack = HeliStack(cona)
        module.synth(stack)
        if hashicorp_configuration_language:
            unformatted = stack.to_hcl_terraform()['hcl'].encode()
            formatted = check_output(['terraform', 'fmt', '-'], input=unformatted).decode()  # noqa: S603
            (top_directory / relative_path).write_text(
                '# AUTOGENERATED by helicopyter\n\n' + formatted
            )
        else:
            dictionary = stack.to_terraform()
            dictionary['//']['AUTOGENERATED'] = 'by helicopyter'
            if stack.imports:
                dictionary['import'] = [
                    {'to': resource, 'id': import_id}
                    for resource, import_id in stack.imports.items()
                ]
            with (top_directory / relative_path).open('w') as output:
                dump(dictionary, output, indent=4, sort_keys=True)
                output.write('\n')


class Parameters(Tap):
    conas: list[str]  # space-separated COdeNAmes
    hashicorp_configuration_language: bool = False

    def configure(self) -> None:  # noqa: D102
        self.add_argument('conas')  # Make this a positional argument


if __name__ == '__main__':
    args = Parameters().parse_args()
    multisynth(args.conas, hashicorp_configuration_language=args.hashicorp_configuration_language)
